/*============================================================================*/
/* ѹ                                                                   */
/*----------------------------------------------------------------------------*/
/* 1.0 : APIб 2003/06/20                                           */
/* 1.1 : ǥߥѲб 2003/09/26                                      */
/* 1.2 : -eѥ᡼ɲ 2004/03/02                                          */
/*============================================================================*/
   
#include <musashi.h>
#include <stdlib.h>
#include <limits.h>

#include <xtnumberHelp.h>
struct mssComHelp comHelp={
  "xtnumber",     /* ޥ̾       */
  "1.2",          /* С       */
  HELPT,          /* ޥɥȥ */
  HELPS,          /*              */
  HELPE,          /*            */
  HELPR,          /* ȥޥ     */
  HELPA,          /* Ծ         */                                      
  HELPB,          /* ХݡȾ */
  HELPH           /* ۡڡ     */
};  

extern struct mssGlobalVariables mssGV;

int sameSRTfld(struct mssFldRecDbl *frd, MssOptFLD *optSRT, int kb){
  int i;
  int fldNo;

  for(i=0; i<optSRT->flds->cnt; i++){
    fldNo=MssFlds2num(optSRT->flds,i);
    if(kb==1) return(0);
    if(strcmp(*(frd->pnt[frd->new]+fldNo),*(frd->pnt[frd->old]+fldNo)) != 0){
      return(0);
    }
  }
  return(1);
}

int main(int argc, char *argv[]){
/*============================================================================*/
/* ץ                                                       */
/*============================================================================*/
/*----------------------------------------------------------------------------*/
/*                                                                    */
/*----------------------------------------------------------------------------*/
  MssOptKEY optKEY={
    OKEY,   /* ץ󥿥                                             */
    "k",    /* (ʣʸԲ)                                   */
    0,      /* 0:ץ, 1:ɬ, 2:XMLtableǤΤɬ(txtǤ̵)      */
    MssFieldMaxCnt, /* ǽʺܿ                                 */
    "i",    /* оݤȤϥǡΥ(GUI)                  */
    2,      /* ǥե(Υץ󤬻ꤵʤäȤư) */
            /* 1:ƤιԤۤ륭ͤȤư                             */
            /* 2:ƤιԤƱͤȤư)                            */
    KEYT,   /* ΥץΥȥ(Helpɽ)                         */
    KEYC    /* ΥץΥ(Helpɽ)                         */
  };

/*----------------------------------------------------------------------------*/
/* ȹ                                                                 */
/*----------------------------------------------------------------------------*/
  MssOptFLD optSRT={
    OFLD,   /* ץ󥿥                                             */
    "s",    /* (ʣʸԲ)                                   */
    0,      /* 0:ץ, 1:ɬ, 2:XMLtableǤΤɬ(txtǤ̵)      */
    MssFieldMaxCnt, /* ǽʺܿ                                 */
    "i",    /* оݤȤϥǡΥ(GUI)                  */
    0,      /* ɽĤ뤫ɤ(0:Բ,1:)                      */
    0,      /* ̾Ǥ뤫ɤ(0:Բ,1:)                    */
    "n,r",  /* ܥץ(%ʲ)ǻǽʸ                        */
            /* ex) ԲĤξNULL, "nr": "-f ̾%rn"λǽ     */
    SRTT,   /* ΥץΥȥ(Helpɽ)                         */
    SRTC,   /* ΥץΥ(Helpɽ)                         */
    SRTF    /* ե饰ˤĤƤ(Helpɽ)ʣξϥޤǶڤ   */
  };

/*----------------------------------------------------------------------------*/
/* ̾                                                                   */
/*----------------------------------------------------------------------------*/
  MssOptSLS optFNM={
    OSLS,   /* ץ󥿥                                             */
    "a",    /* (ʣʸԲ)                                   */
    2,      /* 0:ץ, 1:ɬ, 2:XMLtableǤΤɬ(txtǤ̵)      */
    NULL,   /* ǥե(ʸ)                                           */
    1,      /* ޤǶڤǿκ                             */
    1,      /* ǤʸĹκǾ                                     */
    MssFieldNameMaxLen,/* ǤʸĹκ                          */
    0,      /* 1:Ǥ˥Ǥ,0:Բ  ex) aaaa:xxxxx            */
    FNMT,   /* ΥץΥȥ(Helpɽ)                         */
    FNMC    /* ΥץΥ(Helpɽ)                         */
  }; 

/*----------------------------------------------------------------------------*/
/* ֹ                                                                   */
/*----------------------------------------------------------------------------*/
  MssOptINT optBGN={
    OINT,   /* ץ󥿥                                             */
    "S",    /* (ʣʸԲ)                                   */
    0,      /* 0:ץ, 1:ɬ, 2:XMLtableǤΤɬ(txtǤ̵)      */
    1,      /* ǥե(ͤȤƻ)                                   */
    INT_MIN,/* Ǿ                                                       */
    INT_MAX,/*                                                        */
    BGNT,   /* ΥץΥȥ(Helpɽ)                         */
    BGNC    /* ΥץΥ(Helpɽ)                         */
  };

/*----------------------------------------------------------------------------*/
/* 󥿡Х                                                               */
/*----------------------------------------------------------------------------*/
  MssOptINT optINT={
    OINT,   /* ץ󥿥                                             */
    "I",    /* (ʣʸԲ)                                   */
    0,      /* 0:ץ, 1:ɬ, 2:XMLtableǤΤɬ(txtǤ̵)      */
    1,      /* ǥե(ͤȤƻ)                                   */
    INT_MIN,/* Ǿ                                                       */
    INT_MAX,/*                                                        */
    INTT,   /* ΥץΥȥ(Helpɽ)                         */
    INTC    /* ΥץΥ(Helpɽ)                         */
  };

/*----------------------------------------------------------------------------*/
/* ƱȹܤͤФ                                             */
/*----------------------------------------------------------------------------*/
  MssOptSEL optSAM={
    OSEL,   /* ץ󥿥                                             */
    "e",    /* (ʣʸԲ)                                   */
    0,      /* 0:ץ, 1:ɬ, 2:XMLtableǤΤɬ(txtǤ̵)      */
    "seq",  /* ǥե(ʸˤ)                                       */
    "seq,same,skip",
            /* ǽͥꥹ                                           */
    SAMT,   /* ΥץΥȥ(Helpɽ)                         */
    SAMC,   /* ΥץΥ(Helpɽ)                         */
    SAMS    /* ե饰ˤĤƤ(Helpɽ)ʣξϥޤǶڤ   */
  };

/*----------------------------------------------------------------------------*/
/* Ϣ                                                                 */
/*----------------------------------------------------------------------------*/
  MssOptFLG optBKY={
    OFLG,   /* ץ󥿥                                             */
    "B",    /* (ʣʸԲ)                                   */
    0,      /* ǥե(Ūˤ0) onˤȤ1ˤ          */
    BKYT,   /* ΥץΥȥ(Helpɽ)                         */
    BKYC    /* ΥץΥ(Helpɽ)                         */
  };

/*----------------------------------------------------------------------------*/
/* 󥷥                                                         */
/*----------------------------------------------------------------------------*/
  MssOptFLG optSEQ={
    OFLG,   /* ץ󥿥                                             */
    "q",    /* (ʣʸԲ)                                   */
    0,      /* ǥե(Ūˤ0) onˤȤ1ˤ          */
    SEQT,   /* ΥץΥȥ(Helpɽ)                         */
    SEQC    /* ΥץΥ(Helpɽ)                         */
  };
    
/*----------------------------------------------------------------------------*/
/* ϥե                                                               */
/*----------------------------------------------------------------------------*/
  MssOptINF optINF={                                                             
    OINF,   /* ץ󥿥                                             */
    "i",    /* (ʣʸԲ)                                   */
    0,      /* 0:ץ, 1:ɬ                                         */
    1,      /* ǽκե                                     */
    0,      /*1:file not foundΥ顼ǽλʤ 0:                   */
    INFT,   /* ΥץΥȥ(Helpɽ)                         */
    INFC    /* ΥץΥ(Helpɽ)                         */
  };

/*----------------------------------------------------------------------------*/
/* ϥե                                                               */
/*----------------------------------------------------------------------------*/
  MssOptOTF optOTF={
    OOTF,   /* ץ󥿥                                             */
    "o",    /* (ʣʸԲ)                                   */
    0,      /* 0:ץ, 1:ɬ                                         */
    OTFT,   /* ΥץΥȥ(Helpɽ)                         */
    OTFC    /* ΥץΥ(Helpɽ)                         */
  };

/*----------------------------------------------------------------------------*/
/* ̽                                                                   */
/*----------------------------------------------------------------------------*/
  MssOptFLG optZIP={
    OFLG,   /* ץ󥿥                                             */
    "z",    /* (ʣʸԲ)                                   */
    0,      /* ǥե(Ūˤ0) onˤȤ1ˤ          */
    ZIPT,   /* ΥץΥȥ(Helpɽ)                         */
    ZIPC    /* ΥץΥ(Helpɽ)                         */
  };

/*----------------------------------------------------------------------------*/
/* plain text                                                                 */
/*----------------------------------------------------------------------------*/
  MssOptFLG optTXT={
    OFLG,   /* ץ󥿥                                             */
    "t",    /* (ʣʸԲ)                                   */
    0,      /* ǥե(Ūˤ0) onˤȤ1ˤ          */
    TXTT,   /* ΥץΥȥ(Helpɽ)                         */
    TXTC    /* ΥץΥ(Helpɽ)                         */
  };

/*----------------------------------------------------------------------------*/
/* եѥǥ쥯ȥ̾                                             */
/*----------------------------------------------------------------------------*/
  MssOptSTR optTMP={
    OSTR,   /* ץ󥿥                                             */
    "T",    /* (ʣʸԲ)                                   */
    0,      /* 0:ץ, 1:ɬ, 2:XMLtableǤΤɬ(txtǤ̵)      */
    MssTempDir, /* ǥե                                               */
    1,      /* ʸκǾĹ                                               */
    MssFileNameMaxLen,  /* ʸκĹ                                   */
    TMPT,   /* ΥץΥȥ(Helpɽ)                         */
    TMPC    /* ΥץΥ(Helpɽ)                         */
  };

/*----------------------------------------------------------------------------*/
/* ץޤȤ                                                       */
/*----------------------------------------------------------------------------*/
  void *opt[]={&optKEY,&optSRT,&optBGN,&optINT,&optSAM,&optBKY,&optFNM,&optSEQ,
               &optINF,&optOTF,&optZIP,&optTXT,&optTMP,NULL};

/*============================================================================*/
/* ѿ                                                             */
/*============================================================================*/
  struct mssHeader    *hdi; /*ϥե<head>Ǽ¤*/
  struct mssHeader    *hdo; /*ϥե<head>Ǽ¤*/
  struct mssFPR       *fpr; /*ϥե빽¤                */
  struct mssFPW       *fpw; /*ϥե빽¤                */
  struct mssFldRecDbl *frd; /*-ԥХåե¤           */
  struct mssFields    *sf;  /*ȹܹ¤                  */
  int sorted;               /*Ⱥѥå                */     

  int seqCnt;             /*⤷ϥη*/
  int samCnt;             /*⤷ϥη(Ʊ̹θ)*/
  int same;
  int skip;
  int kb;

/*----------------------------------------------------------------------------*/
/*                                                                      */
/*----------------------------------------------------------------------------*/
  mssInit(argc,argv,&comHelp);       /* ʥʤɤν              */
  mssHelpDoc(opt,&comHelp,argc,argv);/* إ                                */
  mssSetOption(opt,argc,argv);       /* ޥɥץ              */
  fpr=mssOpenFPR(optINF.str,4);      /* ϥե륪ץ                  */
  hdi=mssReadHeader(fpr);            /* إåɤ߹                      */
  mssSetOptKey(&optKEY, hdi);        /* -k ܤإåܤ˴ϢŤ     */
  mssSetOptFld(&optSRT, hdi);        /* -s ܤإåܤ˴ϢŤ     */
    
  /*ȹܤκ*/
  sf=mssInitFields();
  mssAddFieldsByFields(sf,optKEY.flds); /* -k ܤ򥽡ȹܤȤƥå    */
  mssAddFieldsByFields(sf,optSRT.flds); /* -s ܤ򥽡ȹܤȤƥå    */
  mssSetFieldsSortPriority(sf);         /* ֹͥϿˤդ   */
  sorted=mssChkSorted(sf,hdi);          /* ȺѤå                 */
  
/*mssShowFields(sf);*/
/*mssShowOption(opt);*/
/*mssShowHeader(hdi);*/

  /*----------------------------*/
  /*ѥ᡼֤ΰ¸å*/
  /*----------------------------*/
  /*-e same|skip  -BƱԲ*/
  if( 0!=strcmp(optSAM.str,"seq") && optBKY.set ){
    mssShowErrMsg("[-e same|skip] cannot be specified with -B");
    exit(mssErrorNoDefault);
  }

  /*-e same|skip -sλ꤬ʤФʤʤ*/
  if( 0!=strcmp(optSAM.str,"seq") && !optSRT.set ){
    mssShowErrMsg("[-e same|skip] must be specified with -s");
    exit(mssErrorNoDefault);
  }

  /*-s  -BƱԲ*/
  if( optSRT.set && optBKY.set ){
    mssShowErrMsg("-s cannot be specified with -B");
    exit(mssErrorNoDefault);
  }


/*----------------------------------------------------------------------------*/
/*ϥإåκȽ                                                    */
/*----------------------------------------------------------------------------*/
  /*ϥإåν(ȥΥԡ)*/
  hdo=mssInitCpyHeader(hdi);

  /*ϥإåܤɲ*/
  mssAddFieldsByFields(hdo->flds,hdi->flds);

  /*̾ɲä*/
  mssAddFieldsByStrList(hdo->flds,optFNM.strList,optFNM.cnt);

  /*ȤɬפʤsfΥȾȿ*/
  if(!optSEQ.set && !sorted){
    mssSetFieldsSort(hdo->flds,sf);
  }
  
  /*ɸϥץ+إåν*/
  fpw=mssOpenFPW(optOTF.str,optZIP.set,0);
  mssWriteHeader(hdo, fpw);

/*------------------------------------------------------------------------------
ᥤ롼
  SeqפǤʤ and sortedǤʤ -> presort + ֥졼
  ¾                         ->           ֥졼
------------------------------------------------------------------------------*/
  same=0;
  skip=0;
  if( 0==strcmp(optSAM.str,"same") ){
    same=1;
    skip=0;
  }

  if( 0==strcmp(optSAM.str,"skip") ){
    same=1;
    skip=1;
  }

  /*ȤɬѤʤХȤƥȺѤߥեȤƥץ*/
  if( !optSEQ.set && !sorted)
    fpr=mssReopenFPRsort(fpr,4,sf,hdi->flds->cnt,optTMP.str);

  frd=mssInitFRD(hdi->flds->cnt);
  samCnt=optBGN.val-optINT.val;
  seqCnt=optBGN.val-optINT.val;
  if(optBKY.set){
    seqCnt += optINT.val;
  }
  kb=1;
  while( EOF!=mssReadFRD(fpr,frd) ){

    /*֥쥤ν*/
    if(mssKeyBreak(frd, &optKEY)){
      if(frd->eof) break;

      if(optBKY.set){
        seqCnt += optINT.val;
      }else{
        samCnt=optBGN.val-optINT.val;
        seqCnt=optBGN.val-optINT.val;
      }
      kb=1;
    }

    mssGV.inCnt++;

    /*̾Ԥν(ȥå)*/
    if(!optBKY.set){
      if(same){
        if( sameSRTfld(frd,&optSRT,kb) ){
          seqCnt += optINT.val;
        }else{
          if(skip){
            seqCnt += optINT.val;
            samCnt = seqCnt;
          }else{
            seqCnt += optINT.val;
            samCnt += optINT.val;
          }
        } 
      /* -e seq ξ */
      }else{
        seqCnt += optINT.val;
      }
    }

    /*񤭽Ф*/
    mssWriteFld(frd->pnt[frd->new],frd->fldCnt,MssFieldDelimStr,fpw);
    if(same){
      mssWriteInt(samCnt,fpw); mssWriteRet(fpw);
    }else{
      mssWriteInt(seqCnt,fpw); mssWriteRet(fpw);
    }

    kb=0;
    mssGV.outCnt++;
  }
  mssFreeFRD(frd);

/*------------------------------------------------------------------------------
եå&λ
------------------------------------------------------------------------------*/
/*printf("allocCnt=%d\n",getAllocCnt());*/
  mssWriteFooter(fpw);    /* եåν             */
  mssCloseFPR(fpr);       /* ϥեΥ     */
  mssCloseFPW(fpw);       /* ϥեΥ     */
  mssFreeFields(sf);      /* ȹܹ¤Τΰ賫 */
  mssFreeHeader(hdi);     /* ϥإåΰ賫         */
  mssFreeHeader(hdo);     /* ϥإåΰ賫         */
  mssFreeOption(opt);     /* ץΰ賫         */
  mssShowEndMsg();        /* λå             */
  mssEnd(mssExitSuccess); /* λ                       */
  return(0);              /* to avoid warning message   */
}
